# Apple's new arm64 architecture creates all sorts of Docker issues. There is support for emulation of amd64
# processors, but it can be a bit of a pain.

# First, you need to have the following environment variable set:
# export DOCKER_DEFAULT_PLATFORM=linux/amd64

# You'll also need the latest version of Docker Desktop: open Settings… > Features in development, then check the box
# to "Use Rosetta for x86/amd64 emulation on Apple Silicon"

# At this point, all your docker builds should be for linux/amd64 (what AWS ECS uses) - slightly slower, but more
# reliable.

# https://stackoverflow.com/questions/65612411/forcing-docker-to-use-linux-amd64-platform-by-default-on-macos

# This command allows you to inspect what architecture an image was built for:
# docker inspect --format='{{.Os}}/{{.Architecture}}' op-dev-service


import contextlib
import docker
import json
import os
import re
import requests
import shutil
import signal
import socket
import subprocess
import sys
import time

from datetime import datetime, timezone
from http.server import BaseHTTPRequestHandler, HTTPServer
from op_service_client import OPServiceClient
from tempfile import TemporaryDirectory
from threading import Thread

from app.utils import op_dev_service, OPLogFormatter, parse_log_msg


os.environ['DOCKER_DEFAULT_PLATFORM'] = 'linux/amd64'


# There are two ways to run an op-dev service:
# 1. Using flask directly (specify the '--soft' flag)
#    - Much simpler
#    - Easier to test local library packages in this context (although also dangerous by the same token)
# 2. Wrapping in a docker container (default)
#    - More complex (dependency on docker)
#    - Can run exactly as it is run when deployed to a URL (no potential issues with requirements resolution, wsgi setup)
#    - Takes slightly longer

# Here are some useful commands if you run into issues with lingering containers:
# docker stop $(docker ps -a -q)
# docker rm $(docker ps -a -q)

# Here's a one-liner to kill any process running on port 5000:
# lsof -t -i tcp:5000 | xargs kill


RUN_SOFT_SCRIPT = '''
import asyncio
import json
import logging
import uvicorn

from op_service.logging_config import configure_logging
from op_service.op_version_metadata import op_version_metadata

async def main():
  configure_logging()
  from app import service
  logging.info('Initialized process: ' + json.dumps(op_version_metadata))
  return service.starlette

app = asyncio.run(main())


# TODO: I don't know what gives, maybe I have a local dependency installed that's out of date or some shit, but any
# communication via SSL inside the asyncio loop seems to fail without the following:
import ssl
ssl._create_default_https_context = ssl._create_unverified_context


# TODO: try this
# It's possible that the service started some async tasks when it was imported
# loop = asyncio.get_event_loop()
# loop.run_until_complete(asyncio.gather(*asyncio.all_tasks()))

# uvicorn app:starlette --host 0.0.0.0 --port 5000 --workers 1 --no-access-log

uvicorn.run(app, host='0.0.0.0', port=%(port)s, access_log=False)
'''


# This should NOT be kept separately - it is duplicated in the op-dev-manager-service package's build code for services
DEFAULT_DOCKERFILE = '''
FROM opdevio/op-dev-service

# Install the requirements first to take advantage of caching
COPY requirements.txt /var/www/
RUN pip install -r /var/www/requirements.txt

# Only copy over the 'app' directory
COPY app /var/www/app

COPY config.json /var/www/.op-dev/config.json
COPY metadata.json /var/www/.op-dev/metadata.json
'''


# https://docker-py.readthedocs.io/en/stable/api.html#module-docker.api.build
def _docker_build(build_context_dir, no_cache=False):
  build_output = ''
  try:
    docker_client = docker.DockerClient()
    stream = docker_client.api.build(path=build_context_dir, platform='linux/amd64', rm=True, nocache=no_cache)
    image_id = None
    for chunk in stream:
      for line in chunk.splitlines():
        status = json.loads(line)
        if 'error' in status:
          raise Exception('Error building docker image: %s' % status['error'])
        if 'stream' in status:
          msg = status['stream']
          build_output += msg

          # Note: this seems incredibly fragile, but it appears to be the best way to get the image_id...
          # https://github.com/docker/docker-py/blob/39bb78ac694d8d6e53882d3dbc9ebc3c92f5519d/docker/models/images.py#L176
          match = re.search(r'(^Successfully built |sha256:)([0-9a-f]+)$', msg)
          if match:
            image_id = match.group(2)
  except:
    # Only print the build output if we failed - should be so fast most times it's not worth flashing over the screen.
    # It also won't show up in production, so let's try to mimic that environment as much as possible.
    print(' \033[31mFAILED\033[0m')
    build_output and print(build_output)
    raise

  if not image_id:
    print(build_output)
    raise Exception('Error building docker image: could not determine image ID from build output')

  return image_id


def _shutdown(container):
  print('\r\033[36mShutting down...\033[0m', end='', flush=True)
  with contextlib.suppress(Exception):
    container.kill()
    container.remove(force=True)
  print('\r', end='', flush=True)
  sys.exit()


def _check_port(port):
  with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
    if s.connect_ex(('localhost', port)) == 0:
      # Port is already in use
      docker_client = docker.DockerClient()
      for container in docker_client.containers.list():
        for port_list in container.ports.values():
          if str(port) in [ x.get('HostPort') for x in (port_list or []) ]:
            cmd = 'docker kill %s' % container.id
            msg = 'A different docker container is already running on port %i. Try a different port using --port PORT, or kill that container using this command: %s' % (port, cmd)
            sys.exit(msg)
      sys.exit('Port %i is already in use. Try a different port using --port PORT.' % port)


def _get_next_open_port(from_port, max_n=10):
  for port in range(from_port, from_port + max_n):
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
      if s.connect_ex(('localhost', port)) != 0:
        return port
  sys.exit(f'No open port in range found (checked ports {from_port} - {from_port + max_n}). Try manually specifying a port using --port PORT.')


def _get_repository_url(pkg_dir):
  # Returns None if the URL cannot be determined (this is not a git package or it has some weird configuration)
  cmd = [ 'git', 'config', '--local', 'remote.origin.url' ]
  repository_url = subprocess.run(cmd, capture_output=True, text=True, cwd=pkg_dir).stdout.strip()
  return repository_url or None


def _get_pkg_id(pkg_dir):
  # Returns None if the package ID cannot be determined (this is not a git package or it has some weird configuration)
  cmd = [ 'git', 'config', '--local', 'op-dev.pkg-id' ]
  pkg_id = subprocess.run(cmd, capture_output=True, text=True, cwd=pkg_dir).stdout.strip()
  return pkg_id or None


# Returns a tuple of the nullable config ID and the deserizlied config dictionary
def _create_config(config_id, config_path, config_namespace, repository_url):
  if config_path:
    with open(config_path, 'r') as file:
      return None, file.read()

  config = None
  if config_id:
    response = op_dev_service.call('get-config', { 'config_id': config_id })
    config = response['value']
  elif repository_url:
      params = { 'repository_url': repository_url, 'config_namespace': config_namespace }
      # TODO: this is overly general - we should catch a specific exception (namespace not found) and throw on others
      with contextlib.suppress(Exception):
        response = op_dev_service.call('get-latest-config', params)
        config = response['config']['value']
        config_id = response['config_id']
  
  if not config:
    print("\rNo config found for namespace '%s' (using empty)" % config_namespace, file=sys.stderr)
    config = '{}'

  return config_id, json.loads(config)


def _create_metadata(port, pkg_id, commit_id, config_id):
  return {
    'domain': 'http://localhost:%i' % port,
    'pkg_id': pkg_id,
    'commit_id': commit_id,
    'config_id': config_id
  }


# Here is some code you were experimenting with that could bring up an HTTP server locally
# class RequestHandler(BaseHTTPRequestHandler):
#
#   def do_GET(self):
#     print(self.client_address)
#     print(self.path)
#     print(self.headers)
#     self.send_response(200, 'OK')
#     self.send_header('Content-type', 'html')
#     self.end_headers()
#
# print('Starting server...')
# HTTPServer(('', 5001), RequestHandler).serve_forever()


def _is_web_client(pkg_dir):
  package_json = os.path.join(pkg_dir, 'package.json')
  return os.path.exists(package_json)


def run(parser, args):
  as_main = args.main is not None

  # Make sure our desired port is not currently in use
  if args.port:
    port = int(args.port)
    _check_port(port)
  else:
    from_port = 8080 if _is_web_client(args.pkg_dir) else 5000
    port = _get_next_open_port(from_port)

  with TemporaryDirectory() as tmp_dir:

    as_main or print('\033[36mGetting configuration...\033[0m', end='', flush=True)
    pkg_id = _get_pkg_id(args.pkg_dir)
    repository_url = _get_repository_url(args.pkg_dir)
    pkg_name = repository_url.rsplit('/', 1)[1]
    config_id, config = _create_config(args.config_id, args.config_path, args.config_namespace, repository_url)

    if _is_web_client(args.pkg_dir):
      as_main or print('\033[2K\r', end='', flush=True)

      if as_main:
        # TODO: you should probably be able to run this way too
        sys.exit("'--main' option is only for service packages")

      env = os.environ.copy()
      env['OP_DEV_CONFIG'] = json.dumps(config, separators=(',', ':'))
      subprocess.check_call([ 'npm', 'install' ], cwd=args.pkg_dir, env=env)

      # This call hangs until explicitly terminated (for hot-reloading)
      # Note that by default, this command would find the next open port (from 8080), but "explicit is better than
      # implicit"...
      subprocess.check_call([ 'npm', 'run', 'serve', '--', '--port', str(port) ], cwd=args.pkg_dir, env=env)

    else:
      metadata = _create_metadata(port, pkg_id, None, config_id)

      volumes = {}
      work_dir = None
      if as_main:
        volumes[ os.getcwd() ] = { 'bind': '/main', 'mode': 'rw' }
        work_dir = '/main'

      as_main or print('\033[2K\r\033[36mBuilding...\033[0m', end='', flush=True)
      image_id = _docker_build(args.pkg_dir, no_cache=args.no_cache)
      as_main or print('\033[2K\r', end='', flush=True)

      env = {
        'OP_DEV_SERVICE_TEST': '1',
        'OP_DEV_CONFIG': json.dumps(config, separators=(',', ':')),
        'OP_DEV_METADATA': json.dumps(metadata, separators=(',', ':'))
      }

      if as_main:
        env['OP_DEV_SERVICE_MAIN'] = json.dumps(args.main, separators=(',', ':'))

      # Now we can run the container
      docker_client = docker.DockerClient()
      container = docker_client.containers.run(
        image_id,
        # command=None if args.main is None else [ 'python', '-B', '-m', 'app', *args.main ],
        name=pkg_name,
        detach=True,
        remove=True,
        ports={ 5000: port },
        environment=env,
        volumes=volumes,
        working_dir=work_dir
      )
      stream = container.logs(stream=True)

      # Register signal handlers so we can clean up the container correctly
      signal.signal(signal.SIGINT, lambda *args: _shutdown(container))
      signal.signal(signal.SIGTERM, lambda *args: _shutdown(container))
      signal.signal(signal.SIGPIPE, lambda *args: _shutdown(container))


    # Note: if we wanted to print a single status message indicating the server was up and listening to requests, we
    # could do so here by starting a separate thread that waits for the health check to return before printing
    # def health_check():
    #   while True:
    #     try:
    #       requests.get('http://localhost:%s/' % port)
    #     except:
    #       time.sleep(.1)
    #     else:
    #       break
    #   print(f'\033[36mRunning on http://localhost:{port} (CTRL + C to quit)\033[0m')
    # Thread(target=health_check).start()

    # Apply our log formatter to the stream output
    formatter = OPLogFormatter(time_format='%-I:%M:%S %p')
    line = ''
    for encoded_chunk in stream:
      # The stream can deliver partial lines, but we do rely on the behavior that a raw newline character cannot be
      # contained within a chunk.
      chunk = encoded_chunk.decode('utf-8')

      # There is an issue here where access logs can get interspersed in the middle of large chunks. We rely on the fact
      # that these access log chunks are small, and thus likely to be contained within one chunk. This is pretty fragile -
      # a better solution would be to split up the logging to avoid writing to just one stream. We may have issues if
      # multiple workers are generating lots of large log messages at once as well, which this workaround will not solve.
      # Not too big an issue, as the only consequence is poorly-formatted log messages.
      # if chunk.startswith('[OP-ACCESS]'):
      #   print(formatter.format(datetime.now(timezone.utc), parse_log_msg(chunk[:-1])))
      #   continue

      line += chunk
      if line.endswith('\n'):
        parsed = parse_log_msg(line[:-1])
        if as_main:
          if parsed.get('s') in (None, 'app/__main__.py'):
            # Only print messages that were called using print or using the logging module from the __main__ file
            print(parsed['m'])
        else:
          print(formatter.format(datetime.now(timezone.utc), parsed))
        line = ''



import json
import os


CONFIG_FILE_PATH = os.path.expanduser('~/.op-dev-config')


def get_config():
  """
  Returns the deserialized CLI config, or an empty dictionary if there was any error
  """
  
  try:
    with open(CONFIG_FILE_PATH) as file:
      serialized_config = file.read()
    return json.loads(serialized_config)
  except FileNotFoundError:
    return {}


def _update_config(config):
  serialized_config = json.dumps(config, indent=2)
  with open(CONFIG_FILE_PATH, 'w') as file:
    file.write(serialized_config)


def run(parser, args):
  secret_group = (args.secret, args.secret_id)
  if any(secret_group) and not all(secret_group):
    parser.error("must specify '--secret-id' and '--secret' together")

  config = get_config()

  update = [ i for i in vars(args) if i != 'cmd' and getattr(args, i) ]
  if update:
    for key in update:
      value = getattr(args, key)
      config[key] = value
    _update_config(config)

  print(json.dumps(config, indent=2))



import git
import os
import shlex
import subprocess

from urllib.parse import urlparse

from app.utils import op_dev_service


CLONE_REPOSITORY_SH = """
set -e
cd %(dest_dir)s
git init
git config --local op-dev.pkg-id '%(pkg_id)s'
git config --local user.name %(name)s
git config --local user.email %(email)s
git config --local credential.helper ''
git config --local --add credential.helper '!op-dev credential-helper --team-id %(team_id)s $@'
git config --local credential.UseHttpPath 'true'
git remote add origin %(repository_url)s
git fetch
git pull origin main
git push --set-upstream origin main
"""

def clone_repository(name, email, repository_url, team_id, pkg_id, dest_dir):
  """
  Clones the repository from the remote URL and initializes to be used with the specified AWS profile. Fails
  if the 'dest_dir' already exists or the AWS profile is invalid.
  """
  os.makedirs(dest_dir)
  cmd = CLONE_REPOSITORY_SH % {
    "pkg_id": pkg_id,
    "dest_dir": shlex.quote(dest_dir),
    "name": shlex.quote(name),
    "email": shlex.quote(email),
    "repository_url": shlex.quote(repository_url),
    "team_id": team_id
  }
  subprocess.check_output(cmd, shell=True)


def run(parser, args):
  user = op_dev_service.call('get-user', None)
  if args.team:
    team = args.team
    team_id = op_dev_service.call('find-team', { 'name': team })
  else:
    teams = op_dev_service.call('list-teams', None)
    if len(teams) > 1:
      parser.error('--team is required when you are a member of multiple teams')
    else:
      team_id = teams[0][0]
  pkg_id, pkg = op_dev_service.call('find-package', { 'team_id': team_id, 'name': args.pkg })

  clone_repository(user['name'], user['email'], pkg['repository_url'], team_id, pkg_id, args.pkg)

  print('Package successfully cloned')



import setuptools
import sys

assert sys.version_info >= (3, 6)

setuptools.setup(
  name='op-dev',
  version='1.0.2',
  author='Alex Exarhos',
  author_email='AlexExarhos@gmail.com',
  packages=[ 'app' ],
  entry_points={
    'console_scripts': [
      'op-dev = app.__main__:main'
    ]
  },
  install_requires=[
    'boto3 >= 1.18.1',
    'op-service-client >= 0.1.0',
    'python-dateutil >= 2.8.1',
    'pytz >= 2021.3',
    'docker >= 4.1.0',
    'gitpython >= 3.0.5',
    'tzlocal >= 2.0.0'
  ]
)



import argparse
import boto3
import json
import os
import pytz
import re
import sys
import time

from datetime import datetime, timezone
from dateutil import parser
from tzlocal import get_localzone

from app.utils import op_dev_service, OPLogFormatter, parse_log_msg


def parse_datetime_str(value, tz):
  d = parser.parse(value)
  if (d.tzinfo is None) or (d.tzinfo.utcoffset(d) is None):
    d = (tz or get_localzone()).localize(d)
  return d


def datetime_to_timestamp(tz_aware_datetime):
  return int((tz_aware_datetime - datetime.fromtimestamp(0, tz=timezone.utc)).total_seconds() * 1000.0)


def timestamp_to_datetime(ms_since_epoch):
  return datetime.fromtimestamp(ms_since_epoch / 1000, tz=timezone.utc)


def eval_filter(filter_fn, request_data):
  try:
    return filter_fn(request_data)
  except:
    return False


def run(parser, args):
  tz = None
  if args.timezone:
    tz = pytz.timezone(args.timezone)

  now = datetime.now(timezone.utc)

  from_datetime = parse_datetime_str(args.from_datetime, tz) if args.from_datetime else now
  if (from_datetime > now):
    sys.exit("The value for '--from' cannot be in the future")

  to_datetime = parse_datetime_str(args.to_datetime, tz) if args.to_datetime else None
  if to_datetime and (to_datetime > now):
    sys.exit("The value for '--to' cannot be in the future (omit it to stream logs)")

  if to_datetime and (to_datetime < from_datetime):
    sys.exit("The value for '--to' cannot come before the value for '--from'")

  time_format = None
  if not args.iso_format:
    # Infer the time format based on the range in question - if we're only looking at a few minutes around now, it's
    # much simpler to read a local time rather than a full ISO string
    today = now.date()
    from_date = from_datetime.astimezone(None).date()
    to_date = to_datetime.astimezone(None).date() if to_datetime else None
    if (from_date == today):
      # Dealing with events from today (presumably we know what day it is...)
      time_format = '%-I:%M:%S %p'
    elif (today - from_date).days < 7:
      # Dealing with events from days in the past week, so the day is sufficient to distinguish them
      time_format = '%A %-I:%M:%S %p'
    else:
      # Dealing with events from beyond the past week, so we need the actual date (but the year is probably overkill)
      time_format = '%-m/%-d %-I:%M:%S %p'

  credentials = op_dev_service.call('get-log-credentials', { 'url': args.url })
  session = boto3.Session(**credentials['aws_config'])
  logs = session.client('logs')

  if args.filter_requests:
    if args.filter_requests == True:
      filter_fn = lambda x: True
    else:
      filter_fn = eval(args.filter_requests)

  formatter = OPLogFormatter(tz=tz, time_format=time_format)
  from_timestamp = datetime_to_timestamp(from_datetime)
  while True:
    kwargs = { 'logGroupName': credentials['log_group'], 'startTime': from_timestamp, 'interleaved': True }
    if to_datetime:
      kwargs['endTime'] = datetime_to_timestamp(to_datetime)
    for response in logs.get_paginator('filter_log_events').paginate(**kwargs):
      for event in response['events']:
        t = event['timestamp']

        # TODO: this is not a perfectly reliable way to do pagination - if two messages have exactly the same
        # timestamp, it's possible that we might miss the latter message if they happen to cross a page boundary
        from_timestamp = (t + 1)

        dt = timestamp_to_datetime(t)
        msg = parse_log_msg(event['message'])

        if args.levels and (msg.get('l') not in args.levels):
          continue

        if args.name and (msg.get('n') not in args.name):
          continue

        if args.service_instance and (msg.get('i') not in args.service_instance):
          continue

        if msg.get('n') == 'op-request':
          try:
            request_data = json.loads(msg['m'])
          except:
            msg['n'] = 'op-request:legacy'

        if args.filter_requests:
          if msg.get('n') != 'op-request':
            continue
          if not eval_filter(filter_fn, { 't': dt, 'si': msg['i'], **request_data }):
            continue

        if args.format_requests:
          if msg.get('n') == 'op-request':
            if request_data['_'] == 0:
              # If we're using a custom formatter, then disregard the "Started request" messages
              continue
            # Make sure we have all the properties available so the filter formatting string doesn't break
            complete_request_data = {
              'c': None,
              'p': None,
              'pv': None,
              'i': None,
              'iv': None,
              'o': None,
              **request_data,
              't': formatter.format_time(dt),
              'si': msg['i']
            }
            print(args.format_requests % complete_request_data)
            continue

        print(formatter.format(dt, msg))

    if to_datetime:
      break

    time.sleep(3)



import argparse
import importlib


parser = argparse.ArgumentParser()
subparsers = parser.add_subparsers(dest='cmd', metavar='{clone,run,test,logs,config}')

# TODO
# This is used to run a lightweight web server that does the following:
# - Vends Python library packages locally so they can be tested in the context of services
# - Exposes information about what services are running locally
# - Allows local config to resolve the correct ports automatically
# daemon = subparsers.add_parser('daemon')
# daemon.add_argument('--port', default='8016', help='override the default port')

run = subparsers.add_parser('run', description='Run a service or web client locally')
run_config = run.add_mutually_exclusive_group()
run_config.add_argument('--config-path', help='path to local configuration set')
run_config.add_argument('--config-id', help='configuration set ID')
run_config.add_argument('--config-namespace', default='local', help='use the latest configuration set from this namespace')
run.add_argument('--pkg-dir', default='.', help='local package directory (defaults to current directory)')
run.add_argument('--port', help='override the default port')
run.add_argument('--no-cache', action='store_true', help='rebuild the docker image without the cache')
run.add_argument(
  '--replace',
  action='store_true',
  help='replace anything running on the target port (by killing the associated process)'
)
run.add_argument(
  '--main',
  nargs=argparse.REMAINDER,
  help='run the package as a module (by invoking its __main__ file). Any successive arguments will be passed as-is. Note that this can only reference files with relative path names inside the current directory (the current directory is mounted a volume in the docker container).'
)
# run.add_argument(
#   '--soft',
#   action='store_true',
#   help='run the service directly using a lightweight server without a docker image - dependencies must be installed '
#        'locally'
# )
# run.add_argument(
#   '--install',
#   nargs='+',
#   help='[NOT YET SUPPORTED, USE --soft FOR NOW] a list of paths to python packages to build and install, overriding '
#        'any version specified in the requirements.txt file. This option is useful if you are developing a library '
#        'package and want to test it in the context of a service. Packages will be built / installed using: pip install '
#        '<pkg_dir>'
# )

TESTS_HELP = '''run the tests from the tst directory with optional restrictions

PATH: only run tests in the given relative file or directory (allows '*' wildcard matches)
CLASS_NAME: only run test classes with this name (allows '*' wildcard matches)
METHOD_NAME: only run test methods with this name (allows '*' wildcard matches)

Examples with a typical project setup:
Run all tests:
  $ op-dev test
  > tst/foo.py:TestA.test_1
  > tst/foo.py:TestA.test_2
  > tst/foo.py:TestB.test
  > tst/bar.py:TestC.test
  > tst/baz/biz.py:TestD.test
Run a specific test:
  $ op-dev test tst/foo.py:TestA.test_1
  > tst/foo.py:TestA.test_1
Run all tests in a file:
  $ op-dev test tst/foo.py
  > tst/foo.py:TestA.test_1
  > tst/foo.py:TestA.test_2
  > tst/foo.py:TestB.test
Run all tests in a file belonging to a certain class:
  $ op-dev test tst/foo.py:TestB
  > tst/foo.py:TestB.test
Run all tests in a file with a certain name belonging to any class:
  $ op-dev test tst/foo.py:*:test
  > tst/foo.py:TestB.test
Run all tests in any file with a certain name belonging to any class:
  $ op-dev test *:*:test
  > tst/foo.py:TestB.test
  > tst/bar.py:TestC.test
  > tst/baz/biz.py:TestD.test
Run multiple specific tests:
  $ op-dev test tst/foo.py:TestB.test tst/bar.py:TestC.test
  > tst/foo.py:TestB.test
  > tst/bar.py:TestC.test
'''

test = subparsers.add_parser('test', description='Test a service', formatter_class=argparse.RawTextHelpFormatter)
test.add_argument('--pkg-dir', default='.', help='local package directory\nDefaults to current directory')
test.add_argument('tests', nargs='*', metavar="PATH[:CLASS_NAME[.METHOD_NAME]]", help=TESTS_HELP)
test.add_argument('--url', required=True, help='URL of running service instance')

FILTER_REQUESTS_HELP = '''a single-argument python lambda function that should return True if the request message should be included in the output and False otherwise. Any exceptions are equivalent to returning False. The lambda is passed a complete request object as input which has the following properties:
- t: datetime
- si: the service instance ID
- r: route name
- ip: source IP address(es) - note that there can be multiple comma / whitespace delimited IPs as this field comes from the AWS X-Forwarded-For header
- ua: source user agent
- c: response code (as int)
- p: profile (unvalidated)
- pv: profile (validated)
- i: input (unvalidated)
- iv: input (validated)
- o: output

Example filters:
"lambda x: x['c'] != 200"
"lambda x: x['ip'].startswith('255.255.255.255')"
"lambda x: x['pv']['user_id'] == '00000000-0000-0000-0000-000000000000'"
"lambda x: x['si'] == 'abcd1234'"
'''

FORMAT_REQUESTS_HELP = '''format string using any of the properties from the complete request object (see --filter-requests). This option will also coalesce the distinct start and end request messages into a single record. Note that some of the available properties, such as 'ip' and 'ua' are not included in the default output.

Example format:
"%%(t)s %%(ip)s %%(ua)s"
'''

logs = subparsers.add_parser('logs', description='Pull logs for a service', formatter_class=argparse.RawTextHelpFormatter)
logs.add_argument('--url', required=True, help='URL of running service instance')
logs.add_argument('--from', dest='from_datetime', help='retrieve logs starting from this date / time (parsed using dateutil.parser)')
logs.add_argument('--to', dest='to_datetime', help='retrieve logs up to this date / time (parsed using dateutil.parser)')
logs.add_argument('--timezone', help="parse arguments and print timestamps in the specified timezone instead of the local one")
logs.add_argument('--levels', nargs='+', metavar='DEBUG|INFO|WARNING|ERROR|CRITICAL', help='if specified, only list messages that match the specified levels')
logs.add_argument('--name', nargs='+', metavar='NAME', help='if specified, only list messages produced by the specified named loggers')
logs.add_argument('--service-instance', nargs='+', metavar='SERVICE_INSTANCE_ID', help='if specified, only list messages produced by the specified service instance IDs')
logs.add_argument('--iso-format', action='store_true', help='print all timestamps in ISO format instead of the abridged versions inferred by the time range in question')
logs.add_argument('--filter-requests', metavar='FILTER', help=FILTER_REQUESTS_HELP)
logs.add_argument('--format-requests', metavar='FORMAT', help=FORMAT_REQUESTS_HELP)

config = subparsers.add_parser('config', description="View / update this utility's configutation")
config.add_argument('--secret-id', help='set the secret ID')
config.add_argument('--secret', help='set the secret')
config.add_argument('--op-dev-api-url', help=argparse.SUPPRESS)

clone = subparsers.add_parser('clone', description='Clone a package and manage the credentials automatically')
clone.add_argument('--pkg', required=True, help='package name to clone (also name of target directory)')
clone.add_argument('--team', help='team name')

# This is used to automatically authenticate our git requests
credential_helper = subparsers.add_parser('credential-helper')
credential_helper.add_argument('--team-id', required=True)
credential_helper.add_argument('operation', action='store', type=str)

# This is used to pull in config files for things that might not be run through the context of this CLI, namely web
# clients at this point. See the 'vue-cli-plugin-op-dev-config' for an example.
config_helper = subparsers.add_parser('config-helper')
config_helper.add_argument('--config-id', required=True)

# This is used to verify that local SSL certificates are installed and configured correctly so that local services can
# be brought up using SSL. Note that it might prompt for sudo access.
cert_helper = subparsers.add_parser('cert-helper')

# TODO: consider adding an 'ssh' command so you can connect to running instances:
# op-dev ssh --url URL --instance-id INSTANCE_ID
# Note that this would likely require some additional infrastructure to register instance IDs...


def main():
  args = parser.parse_args()

  # Check for a missing 'cmd' here instead of setting 'subparsers.required = True', so we can display a more sensible
  # error message (we had to override metavar to hide some subcommands that we don't want to expose to the user)
  if not args.cmd:
    parser.error('the following arguments are required: cmd')

  module_name = args.cmd.replace('-', '_')
  module = importlib.import_module('.' + module_name, package='app')

  try:
    module.run(parser, args)
  except KeyboardInterrupt:
    # Handle ctrl + C gracefully without printing an error message
    pass


if __name__ == '__main__':
  main()


# def get_pkg(args):
#   pkgs = set()
#   if args.latest_from_url:
#     # TODO: look up this service in the op-dev manager
#     pkgs.add('latest_from_url')
#   if args.pkg_dir:
#     # TODO: figure out the pkg name by analyzing the git config in the pkg dir
#     pkgs.add('pkg_dir')
#   if args.pkg:
#     pkgs.add(args.pkg)

#   if not pkg:
#     parser.error('Could not determine package name')

#   if len(pkgs)
#     #connect to this URL and pull the latest deployment's commit / config IDs







"""
This utility is designed to help you build python service packages / libraries

op-dev run [--latest-from-url http://localhost:5000]
op-dev test --url http://localhost:5000
op-dev logs --url http://localhost:5000
"""


# TODO: add tab-based autocompletion
# https://argcomplete.readthedocs.io/en/latest/
